!> @file lpm_extend_tails.f90
!--------------------------------------------------------------------------------!
! This file is part of PALM.
!
! PALM is free software: you can redistribute it and/or modify it under the terms
! of the GNU General Public License as published by the Free Software Foundation,
! either version 3 of the License, or (at your option) any later version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2014 Leibniz Universitaet Hannover
!--------------------------------------------------------------------------------!
!
! Current revisions:
! ------------------
! 
! 
! Former revisions:
! -----------------
! $Id: lpm_extend_tails.f90 1683 2015-10-07 23:57:51Z knoop $
!
! 1682 2015-10-07 23:56:08Z knoop
! Code annotations made doxygen readable 
! 
! 1359 2014-04-11 17:15:14Z hoffmann
! New particle structure integrated.
! Kind definition added to all floating point numbers.
!
! 1320 2014-03-20 08:40:49Z raasch
! ONLY-attribute added to USE-statements,
! kind-parameters added to all INTEGER and REAL declaration statements,
! kinds are defined in new module kinds,
! comment fields (!:) to be used for variable explanations added to
! all variable declaration statements
!
! 1036 2012-10-22 13:43:42Z raasch
! code put under GPL (PALM 3.9)
!
! 849 2012-03-15 10:35:09Z raasch
! initial revision (former part of advec_particles)
!
!
! Description:
! ------------
!> Add the current particle positions to the particle tails.
!------------------------------------------------------------------------------!
 SUBROUTINE lpm_extend_tails
 

    USE control_parameters,                                                    &
        ONLY:  dt_3d

    USE kinds

    USE particle_attributes,                                                   &
        ONLY:  maximum_number_of_tailpoints, maximum_tailpoint_age,            &
               minimum_tailpoint_distance, number_of_particles, particles,     &
               particle_tail_coordinates

    IMPLICIT NONE

    INTEGER(iwp) ::  i       !<
    INTEGER(iwp) ::  n       !<
    INTEGER(iwp) ::  nn      !<

    REAL(wp) ::  distance    !<


    distance = 0.0_wp

    DO  n = 1, number_of_particles

       nn = particles(n)%tail_id

       IF ( nn /= 0 )  THEN
!
!--       Calculate the distance between the actual particle position and the
!--       next tailpoint
          IF ( minimum_tailpoint_distance /= 0.0_wp )  THEN
             distance = ( particle_tail_coordinates(1,1,nn) -      &
                          particle_tail_coordinates(2,1,nn) )**2 + &
                        ( particle_tail_coordinates(1,2,nn) -      &
                          particle_tail_coordinates(2,2,nn) )**2 + &
                        ( particle_tail_coordinates(1,3,nn) -      &
                          particle_tail_coordinates(2,3,nn) )**2
          ENDIF

!
!--       First, increase the index of all existings tailpoints by one
          IF ( distance >= minimum_tailpoint_distance )  THEN

             DO  i = particles(n)%tailpoints, 1, -1
                particle_tail_coordinates(i+1,:,nn) = &
                                               particle_tail_coordinates(i,:,nn)
             ENDDO
!
!--          Increase the counter which contains the number of tailpoints.
!--          This must always be smaller than the given maximum number of
!--          tailpoints because otherwise the index bounds of
!--          particle_tail_coordinates would be exceeded
             IF ( particles(n)%tailpoints < maximum_number_of_tailpoints-1 ) &
             THEN
                particles(n)%tailpoints = particles(n)%tailpoints + 1
             ENDIF
          ENDIF
!
!--       In any case, store the new point at the beginning of the tail
          particle_tail_coordinates(1,1,nn) = particles(n)%x
          particle_tail_coordinates(1,2,nn) = particles(n)%y
          particle_tail_coordinates(1,3,nn) = particles(n)%z
          particle_tail_coordinates(1,4,nn) = particles(n)%class
!
!--       Increase the age of the tailpoints
          IF ( minimum_tailpoint_distance /= 0.0_wp )  THEN
             particle_tail_coordinates(2:particles(n)%tailpoints,5,nn) =    &
               particle_tail_coordinates(2:particles(n)%tailpoints,5,nn) + dt_3d
!
!--          Delete the last tailpoint, if it has exceeded its maximum age
             IF ( particle_tail_coordinates(particles(n)%tailpoints,5,nn) > &
                  maximum_tailpoint_age )  THEN
                particles(n)%tailpoints = particles(n)%tailpoints - 1
             ENDIF
          ENDIF

       ENDIF

    ENDDO


 END SUBROUTINE lpm_extend_tails
